@@ -28,6 +28,7 @@
 #include <fcntl.h>
 #include <arpa/inet.h>
 
+#include "ui.h"
 #include "die.h"
 #include "xmalloc.h"
 #include "conntrack.h"
@@ -62,6 +63,7 @@ struct flow_entry {
 	uint64_t pkts_dst, bytes_dst;
 	uint64_t timestamp_start, timestamp_stop;
 	char country_src[128], country_dst[128];
+	char country_code_src[4], country_code_dst[4];
 	char city_src[128], city_dst[128];
 	char rev_dns_src[256], rev_dns_dst[256];
 	char procname[256];
@@ -138,6 +140,21 @@ static bool resolve_geoip = true;
 static enum rate_units rate_type = RATE_BYTES;
 static bool show_active_only = false;
 
+enum tbl_flow_col {
+	TBL_FLOW_PROCESS,
+	TBL_FLOW_PID,
+	TBL_FLOW_PROTO,
+	TBL_FLOW_STATE,
+	TBL_FLOW_TIME,
+	TBL_FLOW_ADDRESS,
+	TBL_FLOW_PORT,
+	TBL_FLOW_GEO,
+	TBL_FLOW_BYTES,
+	TBL_FLOW_RATE,
+};
+
+static struct ui_table flows_tbl;
+
 static const char *short_options = "vhTUsDIS46ut:nGb";
 static const struct option long_options[] = {
 	{"ipv4",	no_argument,		NULL, '4'},
@@ -166,11 +183,6 @@ static const char *copyright = "Please report bugs to <netsniff-ng@googlegroups.
 	"This is free software: you are free to change and redistribute it.\n"
 	"There is NO WARRANTY, to the extent permitted by law.";
 
-static const char *const l3proto2str[AF_MAX] = {
-	[AF_INET]			= "ipv4",
-	[AF_INET6]			= "ipv6",
-};
-
 static const char *const l4proto2str[IPPROTO_MAX] = {
 	[IPPROTO_TCP]			= "tcp",
 	[IPPROTO_UDP]			= "udp",
@@ -194,40 +206,40 @@ static const char *const l4proto2str[IPPROTO_MAX] = {
 };
 
 static const char *const tcp_state2str[TCP_CONNTRACK_MAX] = {
-	[TCP_CONNTRACK_NONE]		= "NOSTATE",
-	[TCP_CONNTRACK_SYN_SENT]	= "SYN_SENT",
-	[TCP_CONNTRACK_SYN_RECV]	= "SYN_RECV",
+	[TCP_CONNTRACK_NONE]		= "NONE",
+	[TCP_CONNTRACK_SYN_SENT]	= "SYN-SENT",
+	[TCP_CONNTRACK_SYN_RECV]	= "SYN-RECV",
 	[TCP_CONNTRACK_ESTABLISHED]	= "ESTABLISHED",
-	[TCP_CONNTRACK_FIN_WAIT]	= "FIN_WAIT",
-	[TCP_CONNTRACK_CLOSE_WAIT]	= "CLOSE_WAIT",
-	[TCP_CONNTRACK_LAST_ACK]	= "LAST_ACK",
-	[TCP_CONNTRACK_TIME_WAIT]	= "TIME_WAIT",
+	[TCP_CONNTRACK_FIN_WAIT]	= "FIN-WAIT",
+	[TCP_CONNTRACK_CLOSE_WAIT]	= "CLOSE-WAIT",
+	[TCP_CONNTRACK_LAST_ACK]	= "LAST-ACK",
+	[TCP_CONNTRACK_TIME_WAIT]	= "TIME-WAIT",
 	[TCP_CONNTRACK_CLOSE]		= "CLOSE",
-	[TCP_CONNTRACK_SYN_SENT2]	= "SYN_SENT2",
+	[TCP_CONNTRACK_SYN_SENT2]	= "SYN-SENT2",
 };
 
 static const char *const dccp_state2str[DCCP_CONNTRACK_MAX] = {
-	[DCCP_CONNTRACK_NONE]		= "NOSTATE",
+	[DCCP_CONNTRACK_NONE]		= "NONE",
 	[DCCP_CONNTRACK_REQUEST]	= "REQUEST",
 	[DCCP_CONNTRACK_RESPOND]	= "RESPOND",
 	[DCCP_CONNTRACK_PARTOPEN]	= "PARTOPEN",
 	[DCCP_CONNTRACK_OPEN]		= "OPEN",
-	[DCCP_CONNTRACK_CLOSEREQ]	= "CLOSEREQ",
+	[DCCP_CONNTRACK_CLOSEREQ]	= "CLOSE-REQ",
 	[DCCP_CONNTRACK_CLOSING]	= "CLOSING",
-	[DCCP_CONNTRACK_TIMEWAIT]	= "TIMEWAIT",
+	[DCCP_CONNTRACK_TIMEWAIT]	= "TIME-WAIT",
 	[DCCP_CONNTRACK_IGNORE]		= "IGNORE",
 	[DCCP_CONNTRACK_INVALID]	= "INVALID",
 };
 
 static const char *const sctp_state2str[SCTP_CONNTRACK_MAX] = {
-	[SCTP_CONNTRACK_NONE]		= "NOSTATE",
+	[SCTP_CONNTRACK_NONE]		= "NONE",
 	[SCTP_CONNTRACK_CLOSED]		= "CLOSED",
-	[SCTP_CONNTRACK_COOKIE_WAIT]	= "COOKIE_WAIT",
-	[SCTP_CONNTRACK_COOKIE_ECHOED]	= "COOKIE_ECHOED",
+	[SCTP_CONNTRACK_COOKIE_WAIT]	= "COOKIE-WAIT",
+	[SCTP_CONNTRACK_COOKIE_ECHOED]	= "COOKIE-ECHO",
 	[SCTP_CONNTRACK_ESTABLISHED]	= "ESTABLISHED",
-	[SCTP_CONNTRACK_SHUTDOWN_SENT]	= "SHUTDOWN_SENT",
-	[SCTP_CONNTRACK_SHUTDOWN_RECD]	= "SHUTDOWN_RECD",
-	[SCTP_CONNTRACK_SHUTDOWN_ACK_SENT] = "SHUTDOWN_ACK_SENT",
+	[SCTP_CONNTRACK_SHUTDOWN_SENT]	= "SHUTD-SENT",
+	[SCTP_CONNTRACK_SHUTDOWN_RECD]	= "SHUTD-RCVD",
+	[SCTP_CONNTRACK_SHUTDOWN_ACK_SENT] = "SHUTD-ACK",
 };
 
 static const struct nfct_filter_ipv4 filter_ipv4 = {
@@ -684,6 +696,7 @@ flow_entry_geo_country_lookup_generic(struct flow_entry *n,
 	struct sockaddr_in sa4;
 	struct sockaddr_in6 sa6;
 	const char *country = NULL;
+	const char *country_code = NULL;
 
 	switch (n->l3_proto) {
 	default:
@@ -692,11 +705,13 @@ flow_entry_geo_country_lookup_generic(struct flow_entry *n,
 	case AF_INET:
 		flow_entry_get_sain4_obj(n, dir, &sa4);
 		country = geoip4_country_name(&sa4);
+		country_code = geoip4_country_code3_name(&sa4);
 		break;
 
 	case AF_INET6:
 		flow_entry_get_sain6_obj(n, dir, &sa6);
 		country = geoip6_country_name(&sa6);
+		country_code = geoip6_country_code3_name(&sa6);
 		break;
 	}
 
@@ -707,6 +722,14 @@ flow_entry_geo_country_lookup_generic(struct flow_entry *n,
 		        sizeof(n->country_src));
 	else
 		SELFLD(dir, country_src, country_dst)[0] = '\0';
+
+	build_bug_on(sizeof(n->country_code_src) != sizeof(n->country_code_dst));
+
+	if (country_code)
+		strlcpy(SELFLD(dir, country_code_src, country_code_dst),
+			country_code, sizeof(n->country_code_src));
+	else
+		SELFLD(dir, country_code_src, country_code_dst)[0] = '\0';
 }
 
 static void flow_entry_get_extended_geo(struct flow_entry *n,
@@ -777,10 +800,8 @@ static void flow_entry_get_extended(struct flow_entry *n)
 	if (n->flow_id == 0)
 		return;
 
-	if (show_src) {
-		flow_entry_get_extended_revdns(n, FLOW_DIR_SRC);
-		flow_entry_get_extended_geo(n, FLOW_DIR_SRC);
-	}
+	flow_entry_get_extended_revdns(n, FLOW_DIR_SRC);
+	flow_entry_get_extended_geo(n, FLOW_DIR_SRC);
 
 	flow_entry_get_extended_revdns(n, FLOW_DIR_DST);
 	flow_entry_get_extended_geo(n, FLOW_DIR_DST);
@@ -792,36 +813,13 @@ static void flow_entry_get_extended(struct flow_entry *n)
 		walk_processes(n);
 }
 
-static uint16_t presenter_get_port(uint16_t src, uint16_t dst, bool is_tcp)
-{
-	if (src < dst && src < 1024) {
-		return src;
-	} else if (dst < src && dst < 1024) {
-		return dst;
-	} else {
-		const char *tmp1, *tmp2;
-		if (is_tcp) {
-			tmp1 = lookup_port_tcp(src);
-			tmp2 = lookup_port_tcp(dst);
-		} else {
-			tmp1 = lookup_port_udp(src);
-			tmp2 = lookup_port_udp(dst);
-		}
-		if (tmp1 && !tmp2) {
-			return src;
-		} else if (!tmp1 && tmp2) {
-			return dst;
-		} else {
-			if (src < dst)
-				return src;
-			else
-				return dst;
-		}
-	}
-}
-
 static char *bandw2str(double bytes, char *buf, size_t len)
 {
+	if (bytes <= 0) {
+		buf[0] = '\0';
+		return buf;
+	}
+
 	if (bytes > 1000000000.)
 		snprintf(buf, len, "%.1fGB", bytes / 1000000000.);
 	else if (bytes > 1000000.)
@@ -829,7 +827,7 @@ static char *bandw2str(double bytes, char *buf, size_t len)
 	else if (bytes > 1000.)
 		snprintf(buf, len, "%.1fkB", bytes / 1000.);
 	else
-		snprintf(buf, len, "%g bytes", bytes);
+		snprintf(buf, len, "%.0f", bytes);
 
 	return buf;
 }
@@ -837,10 +835,15 @@ static char *bandw2str(double bytes, char *buf, size_t len)
 static char *rate2str(double rate, char *buf, size_t len)
 {
 	const char * const unit_fmt[2][4] = {
-		{ "%.1fGbit/s", "%.1fMbit/s", "%.1fkbit/s", "%gbit/s" },
-		{ "%.1fGB/s",   "%.1fMB/s",   "%.1fkB/s",   "%gB/s"   }
+		{ "%.1fGbit/s", "%.1fMbit/s", "%.1fkbit/s", "%.0fbit/s" },
+		{ "%.1fGB/s",   "%.1fMB/s",   "%.1fkB/s",   "%.0fB/s"   }
 	};
 
+	if (rate <= 0) {
+		buf[0] = '\0';
+		return buf;
+	}
+
 	if (rate_type == RATE_BITS)
 		rate *= 8;
 
@@ -856,176 +859,178 @@ static char *rate2str(double rate, char *buf, size_t len)
 	return buf;
 }
 
-static void presenter_print_counters(uint64_t bytes, uint64_t pkts,
-				     double rate_bytes, double rate_pkts,
-				     int color)
+static char *time2str(uint64_t tstamp, char *str, size_t len)
 {
-	char bytes_str[64];
-
-	printw(" -> (");
-	attron(COLOR_PAIR(color));
-	printw("%"PRIu64" pkts", pkts);
-	if (rate_pkts) {
-		attron(COLOR_PAIR(3));
-		printw("(%.1fpps)", rate_pkts);
-		attron(COLOR_PAIR(color));
-	}
-
-	printw(", %s", bandw2str(bytes, bytes_str, sizeof(bytes_str) - 1));
-	if (rate_bytes) {
-		attron(COLOR_PAIR(3));
-		printw("(%s)", rate2str(rate_bytes, bytes_str,
-			sizeof(bytes_str) - 1));
-		attron(COLOR_PAIR(color));
-	}
-	attroff(COLOR_PAIR(color));
-	printw(")");
-}
-
-static void presenter_print_flow_entry_time(const struct flow_entry *n)
-{
-	int h, m, s;
 	time_t now;
+	int v, s;
 
 	time(&now);
 
-	s = now - (n->timestamp_start / NSEC_PER_SEC);
-	if (s <= 0)
-		return;
-
-	h = s / 3600;
-	s -= h * 3600;
-	m = s / 60;
-	s -= m * 60;
-
-	printw(" [ time");
-	if (h > 0)
-		printw(" %dh", h);
-	if (m > 0)
-		printw(" %dm", m);
-	if (s > 0)
-		printw(" %ds", s);
-	printw(" ]");
-}
-
-static void draw_flow_entry(WINDOW *screen, const struct flow_entry *n,
-			    unsigned int *line)
-{
-	char tmp[128];
-	const char *pname = NULL;
-	uint16_t port;
-
-	mvwprintw(screen, *line, 2, "");
-
-	/* PID, application name */
-	if (n->procnum > 0) {
-		slprintf(tmp, sizeof(tmp), "%s(%d)", n->procname, n->procnum);
-
-		printw("[");
-		attron(COLOR_PAIR(3));
-		printw("%s", tmp);
-		attroff(COLOR_PAIR(3));
-		printw("]:");
+	s = now - (tstamp ? (tstamp / NSEC_PER_SEC) : now);
+	if (s <= 0) {
+		str[0] = '\0';
+		return str;
 	}
 
-	/* L3 protocol, L4 protocol, states */
-	printw("%s:%s", l3proto2str[n->l3_proto], l4proto2str[n->l4_proto]);
-	printw("[");
-	attron(COLOR_PAIR(3));
+	v = s / (3600 * 24);
+	if (v > 0) {
+		slprintf(str, len, "%dd", v);
+		return str;
+	}
+
+	v = s / 3600;
+	if (v > 0) {
+		slprintf(str, len, "%dh", v);
+		return str;
+	}
+
+	v = s / 60;
+	if (v > 0) {
+		slprintf(str, len, "%dm", v);
+		return str;
+	}
+
+	slprintf(str, len, "%ds", s);
+	return str;
+}
+
+
+static const char *flow_state2str(const struct flow_entry *n)
+{
 	switch (n->l4_proto) {
 	case IPPROTO_TCP:
-		printw("%s", tcp_state2str[n->tcp_state]);
-		break;
+		return tcp_state2str[n->tcp_state];
 	case IPPROTO_SCTP:
-		printw("%s", sctp_state2str[n->sctp_state]);
-		break;
+		return sctp_state2str[n->sctp_state];
 	case IPPROTO_DCCP:
-		printw("%s", dccp_state2str[n->dccp_state]);
-		break;
+		return dccp_state2str[n->dccp_state];
+
 	case IPPROTO_UDP:
 	case IPPROTO_UDPLITE:
 	case IPPROTO_ICMP:
 	case IPPROTO_ICMPV6:
-		printw("NOSTATE");
-		break;
+	default:
+		return "";
 	}
-	attroff(COLOR_PAIR(3));
-	printw("]");
+}
+
+static char *flow_port2str(const struct flow_entry *n, char *str, size_t len,
+		           enum flow_direction dir)
+{
+	const char *tmp = NULL;
+	uint16_t port = 0;
+
+	port = SELFLD(dir, port_src, port_dst);
+	tmp = NULL;
 
-	/* Guess application port */
 	switch (n->l4_proto) {
 	case IPPROTO_TCP:
-		port = presenter_get_port(n->port_src, n->port_dst, true);
-		pname = lookup_port_tcp(port);
+		tmp = lookup_port_tcp(port);
 		break;
 	case IPPROTO_UDP:
 	case IPPROTO_UDPLITE:
-		port = presenter_get_port(n->port_src, n->port_dst, false);
-		pname = lookup_port_udp(port);
+		tmp = lookup_port_udp(port);
 		break;
 	}
-	if (pname) {
-		attron(A_BOLD);
-		printw(":%s", pname);
-		attroff(A_BOLD);
+
+	if (!tmp && port)
+		slprintf(str, len, "%d", port);
+	else
+		slprintf(str, len, "%s", tmp ? tmp : "");
+
+	return str;
+}
+
+static void print_flow_peer_info(const struct flow_entry *n, enum flow_direction dir)
+{
+	int counters_color = COLOR(YELLOW, BLACK);
+	int src_color = COLOR(RED, BLACK);
+	int dst_color = COLOR(BLUE, BLACK);
+	int country_color = COLOR(GREEN, BLACK);
+	int addr_color = dst_color;
+	int port_color = A_BOLD;
+	char tmp[128];
+
+	if (show_src && dir == FLOW_DIR_SRC) {
+		country_color  = src_color;
+		counters_color = src_color;
+		port_color    |= src_color;
+		addr_color     = src_color;
+	} else if (show_src && FLOW_DIR_DST) {
+		country_color  = dst_color;
+		counters_color = dst_color;
+		port_color    |= dst_color;
+		addr_color     = dst_color;
 	}
 
-	if (n->timestamp_start > 0)
-		presenter_print_flow_entry_time(n);
+	ui_table_col_color_set(&flows_tbl, TBL_FLOW_ADDRESS, addr_color);
+	ui_table_col_color_set(&flows_tbl, TBL_FLOW_PORT, port_color);
+	ui_table_col_color_set(&flows_tbl, TBL_FLOW_GEO, country_color);
+	ui_table_col_color_set(&flows_tbl, TBL_FLOW_BYTES, counters_color);
+	ui_table_col_color_set(&flows_tbl, TBL_FLOW_RATE, counters_color);
+
+	/* Reverse DNS/IP */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_ADDRESS,
+			      SELFLD(dir, rev_dns_src, rev_dns_dst));
+
+	/* Application port */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_PORT,
+			      flow_port2str(n, tmp, sizeof(tmp), dir));
+
+	/* GEO */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_GEO,
+			      SELFLD(dir, country_code_src, country_code_dst));
+
+	/* Bytes */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_BYTES,
+			      bandw2str(SELFLD(dir, bytes_src, bytes_dst),
+					tmp, sizeof(tmp) - 1));
+
+	/* Rate bytes */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_RATE,
+			      rate2str(SELFLD(dir, rate_bytes_src, rate_bytes_dst),
+				       tmp, sizeof(tmp) - 1));
+}
+
+static void draw_flow_entry(const struct flow_entry *n)
+{
+	char tmp[128];
+
+	ui_table_row_add(&flows_tbl);
+
+	/* Application */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_PROCESS, n->procname);
+
+	/* PID */
+	slprintf(tmp, sizeof(tmp), "%.d", n->procnum);
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_PID, tmp);
+
+	/* L4 protocol */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_PROTO, l4proto2str[n->l4_proto]);
+
+	/* L4 protocol state */
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_STATE, flow_state2str(n));
+
+	/* Time */
+	time2str(n->timestamp_start, tmp, sizeof(tmp));
+	ui_table_row_col_set(&flows_tbl, TBL_FLOW_TIME, tmp);
+
+	print_flow_peer_info(n, show_src ? FLOW_DIR_SRC : FLOW_DIR_DST);
+
+	ui_table_row_show(&flows_tbl);
 
-	/* Show source information: reverse DNS, port, country, city, counters */
 	if (show_src) {
-		attron(COLOR_PAIR(1));
-		mvwprintw(screen, ++(*line), 8, "src: %s", n->rev_dns_src);
-		attroff(COLOR_PAIR(1));
+		ui_table_row_add(&flows_tbl);
 
-		printw(":%"PRIu16, n->port_src);
+		ui_table_row_col_set(&flows_tbl, TBL_FLOW_PROCESS, "");
+		ui_table_row_col_set(&flows_tbl, TBL_FLOW_PID, "");
+		ui_table_row_col_set(&flows_tbl, TBL_FLOW_PROTO, "");
+		ui_table_row_col_set(&flows_tbl, TBL_FLOW_STATE, "");
+		ui_table_row_col_set(&flows_tbl, TBL_FLOW_TIME, "");
 
-		if (n->country_src[0]) {
-			printw(" (");
-
-			attron(COLOR_PAIR(4));
-			printw("%s", n->country_src);
-			attroff(COLOR_PAIR(4));
-
-			if (n->city_src[0])
-				printw(", %s", n->city_src);
-
-			printw(")");
-		}
-
-		if (n->pkts_src > 0 && n->bytes_src > 0)
-			presenter_print_counters(n->bytes_src, n->pkts_src,
-						 n->rate_bytes_src,
-						 n->rate_pkts_src, 1);
-
-		printw(" => ");
+		print_flow_peer_info(n, FLOW_DIR_DST);
+		ui_table_row_show(&flows_tbl);
 	}
-
-	/* Show dest information: reverse DNS, port, country, city, counters */
-	attron(COLOR_PAIR(2));
-	mvwprintw(screen, ++(*line), 8, "dst: %s", n->rev_dns_dst);
-	attroff(COLOR_PAIR(2));
-
-	printw(":%"PRIu16, n->port_dst);
-
-	if (n->country_dst[0]) {
-		printw(" (");
-
-		attron(COLOR_PAIR(4));
-		printw("%s", n->country_dst);
-		attroff(COLOR_PAIR(4));
-
-		if (n->city_dst[0])
-			printw(", %s", n->city_dst);
-
-		printw(")");
-	}
-
-	if (n->pkts_dst > 0 && n->bytes_dst > 0)
-		presenter_print_counters(n->bytes_dst, n->pkts_dst,
-					 n->rate_bytes_dst,
-					 n->rate_pkts_dst, 2);
 }
 
 static inline bool presenter_flow_wrong_state(struct flow_entry *n)
@@ -1091,14 +1096,11 @@ static inline bool presenter_flow_wrong_state(struct flow_entry *n)
 static void draw_flows(WINDOW *screen, struct flow_list *fl,
 		       int skip_lines)
 {
-	int skip_left = skip_lines;
+	int row_width = show_src ? 2 : 1;
 	unsigned int flows = 0;
-	unsigned int line = 3;
+	unsigned int line = 4;
+	int skip = skip_lines;
 	struct flow_entry *n;
-	int maxy = rows - 6;
-
-	wclear(screen);
-	clear();
 
 	rcu_read_lock();
 
@@ -1107,30 +1109,28 @@ static void draw_flows(WINDOW *screen, struct flow_list *fl,
 		mvwprintw(screen, line, 2, "(No sessions! "
 			  "Is netfilter running?)");
 
+	ui_table_clear(&flows_tbl);
+	ui_table_header_print(&flows_tbl);
+
 	for (; n; n = rcu_dereference(n->next)) {
 		if (!n->is_visible)
 			continue;
-
 		if (presenter_flow_wrong_state(n))
 			continue;
 
 		/* count only flows which might be showed */
 		flows++;
 
-		if (maxy <= 0)
+		if (line + row_width >= rows)
+			continue;
+		if (--skip >= 0)
 			continue;
 
-		if (skip_left > 0) {
-			skip_left--;
-			continue;
-		}
-
-		draw_flow_entry(screen, n, &line);
-
-		line++;
-		maxy -= (2 + (show_src ? 1 : 0));
+		draw_flow_entry(n);
+		line += row_width;
 	}
 
+	mvwprintw(screen, 1, 0, "%*s", COLS - 1, " ");
 	mvwprintw(screen, 1, 2, "Kernel netfilter flows(%u) for ", flows);
 
 	if (what & INCLUDE_IPV4)
@@ -1160,7 +1160,7 @@ static void draw_flows(WINDOW *screen, struct flow_list *fl,
 	rcu_read_unlock();
 }
 
-static void draw_help(WINDOW *screen)
+static void draw_help(void)
 {
 	int col = 0;
 	int row = 1;
@@ -1190,21 +1190,24 @@ static void draw_help(WINDOW *screen)
 
 	mvaddnstr(row + 4, col + 3, "Up, u, k      Move up", -1);
 	mvaddnstr(row + 5, col + 3, "Down, d, j    Move down", -1);
-	mvaddnstr(row + 6, col + 3, "?             Toggle help window", -1);
-	mvaddnstr(row + 7, col + 3, "q, Ctrl+C     Quit", -1);
+	mvaddnstr(row + 6, col + 3, "Left,l        Scroll left", -1);
+	mvaddnstr(row + 7, col + 3, "Right,h       Scroll right", -1);
+	mvaddnstr(row + 8, col + 3, "?             Toggle help window", -1);
+	mvaddnstr(row + 9, col + 3, "q, Ctrl+C     Quit", -1);
 
 	attron(A_BOLD | A_UNDERLINE);
-	mvaddnstr(row + 9, col + 2, "Display Settings", -1);
+	mvaddnstr(row + 11, col + 2, "Display Settings", -1);
 	attroff(A_BOLD | A_UNDERLINE);
 
-	mvaddnstr(row + 11, col + 3, "b     Toggle rate units (bits/bytes)", -1);
-	mvaddnstr(row + 12, col + 3, "a     Toggle display of active flows (rate > 0) only", -1);
+	mvaddnstr(row + 13, col + 3, "b     Toggle rate units (bits/bytes)", -1);
+	mvaddnstr(row + 14, col + 3, "a     Toggle display of active flows (rate > 0) only", -1);
+	mvaddnstr(row + 15, col + 3, "s     Toggle show source peer info", -1);
 
-	mvaddnstr(row + 14, col + 3, "T     Toggle display TCP flows", -1);
-	mvaddnstr(row + 15, col + 3, "U     Toggle display UDP flows", -1);
-	mvaddnstr(row + 16, col + 3, "D     Toggle display DCCP flows", -1);
-	mvaddnstr(row + 17, col + 3, "I     Toggle display ICMP flows", -1);
-	mvaddnstr(row + 18, col + 3, "S     Toggle display SCTP flows", -1);
+	mvaddnstr(row + 17, col + 3, "T     Toggle display TCP flows", -1);
+	mvaddnstr(row + 18, col + 3, "U     Toggle display UDP flows", -1);
+	mvaddnstr(row + 19, col + 3, "D     Toggle display DCCP flows", -1);
+	mvaddnstr(row + 20, col + 3, "I     Toggle display ICMP flows", -1);
+	mvaddnstr(row + 21, col + 3, "S     Toggle display SCTP flows", -1);
 }
 
 static void draw_header(WINDOW *screen)
@@ -1220,7 +1223,7 @@ static void draw_header(WINDOW *screen)
 	attroff(A_STANDOUT);
 }
 
-static void draw_footer(WINDOW *screen)
+static void draw_footer(void)
 {
 	int i;
 
@@ -1255,27 +1258,58 @@ static void show_option_toggle(int opt)
 	}
 }
 
+static void flows_table_init(struct ui_table *tbl)
+{
+	ui_table_init(tbl);
+
+	ui_table_pos_set(tbl, 3, 0);
+	ui_table_height_set(tbl, LINES - 3);
+
+	ui_table_col_add(tbl, TBL_FLOW_PROCESS, "PROCESS", 13);
+	ui_table_col_add(tbl, TBL_FLOW_PID, "PID", 7);
+	ui_table_col_add(tbl, TBL_FLOW_PROTO, "PROTO", 6);
+	ui_table_col_add(tbl, TBL_FLOW_STATE, "STATE", 11);
+	ui_table_col_add(tbl, TBL_FLOW_TIME, "TIME", 4);
+	ui_table_col_add(tbl, TBL_FLOW_ADDRESS, "ADDRESS", 50);
+	ui_table_col_add(tbl, TBL_FLOW_PORT, "PORT", 8);
+	ui_table_col_add(tbl, TBL_FLOW_GEO, "GEO", 3);
+	ui_table_col_add(tbl, TBL_FLOW_BYTES, "BYTES", 10);
+	ui_table_col_add(tbl, TBL_FLOW_RATE, "RATE", 10);
+
+	ui_table_col_align_set(tbl, TBL_FLOW_TIME, UI_ALIGN_RIGHT);
+	ui_table_col_align_set(tbl, TBL_FLOW_BYTES, UI_ALIGN_RIGHT);
+	ui_table_col_align_set(tbl, TBL_FLOW_RATE, UI_ALIGN_RIGHT);
+
+	ui_table_col_color_set(tbl, TBL_FLOW_PROCESS, COLOR(YELLOW, BLACK));
+	ui_table_col_color_set(tbl, TBL_FLOW_PID, A_BOLD);
+	ui_table_col_color_set(tbl, TBL_FLOW_STATE, COLOR(YELLOW, BLACK));
+
+	ui_table_header_color_set(&flows_tbl, COLOR(BLACK, GREEN));
+}
+
 static void presenter(void)
 {
-	int time_sleep_us = 200000;
-	int time_passed_us = 0;
 	bool show_help = false;
 	int skip_lines = 0;
 	WINDOW *screen;
 
 	lookup_init(LT_PORTS_TCP);
 	lookup_init(LT_PORTS_UDP);
+
 	screen = screen_init(false);
+	wclear(screen);
 
 	start_color();
-	init_pair(1, COLOR_RED, COLOR_BLACK);
-	init_pair(2, COLOR_BLUE, COLOR_BLACK);
-	init_pair(3, COLOR_YELLOW, COLOR_BLACK);
-	init_pair(4, COLOR_GREEN, COLOR_BLACK);
+	INIT_COLOR(RED, BLACK);
+	INIT_COLOR(BLUE, BLACK);
+	INIT_COLOR(YELLOW, BLACK);
+	INIT_COLOR(GREEN, BLACK);
+	INIT_COLOR(BLACK, GREEN);
+
+        flows_table_init(&flows_tbl);
 
 	rcu_register_thread();
 	while (!sigint) {
-		bool redraw_flows = true;
 		int ch;
 
 		curs_set(0);
@@ -1300,6 +1334,14 @@ static void presenter(void)
 			if (skip_lines > SCROLL_MAX)
 				skip_lines = SCROLL_MAX;
 			break;
+		case KEY_LEFT:
+		case 'h':
+			ui_table_event_send(&flows_tbl, UI_EVT_SCROLL_LEFT);
+			break;
+		case KEY_RIGHT:
+		case 'l':
+			ui_table_event_send(&flows_tbl, UI_EVT_SCROLL_RIGHT);
+			break;
 		case 'b':
 			if (rate_type == RATE_BYTES)
 				rate_type = RATE_BITS;
@@ -1309,6 +1351,9 @@ static void presenter(void)
 		case 'a':
 			show_active_only = !show_active_only;
 			break;
+		case 's':
+			show_src = !show_src;
+			break;
 		case '?':
 			show_help = !show_help;
 			wclear(screen);
@@ -1324,36 +1369,24 @@ static void presenter(void)
 			break;
 		default:
 			fflush(stdin);
-			redraw_flows = false;
 			break;
 		}
 
-		if (!redraw_flows)
-			redraw_flows = time_passed_us >= 1 * USEC_PER_SEC;
-
-		if (show_help)
-			redraw_flows = false;
-
-		if (redraw_flows) {
-			draw_flows(screen, &flow_list, skip_lines);
-			time_passed_us = 0;
-		} else {
-			time_passed_us += time_sleep_us;
-		}
-
 		draw_header(screen);
 
 		if (show_help)
-			draw_help(screen);
+			draw_help();
+		else
+			draw_flows(screen, &flow_list, skip_lines);
 
-		draw_footer(screen);
+		draw_footer();
 
-		wrefresh(screen);
-		refresh();
-		usleep(time_sleep_us);
+		usleep(80000);
 	}
 	rcu_unregister_thread();
 
+	ui_table_uninit(&flows_tbl);
+
 	screen_end();
 	lookup_cleanup(LT_PORTS_UDP);
 	lookup_cleanup(LT_PORTS_TCP);
@@ -1425,7 +1458,7 @@ static int flow_list_update_entry(struct flow_list *fl, struct nf_conntrack *ct)
 {
 	struct flow_entry *n;
 
-	n = flow_list_find_id(&flow_list, nfct_get_attr_u32(ct, ATTR_ID));
+	n = flow_list_find_id(fl, nfct_get_attr_u32(ct, ATTR_ID));
 	if (!n)
 		return NFCT_CB_CONTINUE;
 
@@ -1519,7 +1552,7 @@ static void collector_create_filter(struct nfct_handle *nfct)
 /* This hand-crafted filter looks ugly but it allows to do not
  * flush nfct connections & filter them by user specified filter.
  * May be it is better to replace this one by nfct_cmp. */
-static int flow_dump_cb(enum nf_conntrack_msg_type type,
+static int flow_dump_cb(enum nf_conntrack_msg_type type __maybe_unused,
 			struct nf_conntrack *ct, void *data __maybe_unused)
 {
 	struct flow_entry fl;
