@@ -55,7 +55,7 @@ enum dump_mode {
 
 struct ctx {
 	char *device_in, *device_out, *device_trans, *filter, *prefix;
-	int cpu, rfraw, dump, print_mode, dump_dir, packet_type;
+	int cpu, rfraw, dump, print_mode, dump_dir, packet_type, lo_ifindex;
 	unsigned long kpull, dump_interval, tx_bytes, tx_packets;
 	size_t reserve_size;
 	bool randomize, promiscuous, enforce, jumbo, dump_bpf, hwtimestamp, verbose;
@@ -66,7 +66,7 @@ struct ctx {
 	uint32_t link_type, magic;
 	uint32_t fanout_group, fanout_type;
 	uint64_t pkts_seen, pkts_recvd, pkts_drops;
-	uint64_t pkts_recvd_last, pkts_drops_last;
+	uint64_t pkts_recvd_last, pkts_drops_last, pkts_skipd_last;
 };
 
 static volatile sig_atomic_t sigint = 0, sighup = 0;
@@ -218,10 +218,13 @@ static int update_rx_stats(struct ctx *ctx, int sock, bool is_v3)
 	if (ret)
 		return ret;
 
+	drops += ctx->pkts_skipd_last;
+	ctx->pkts_seen += ctx->pkts_skipd_last;
 	ctx->pkts_recvd += packets;
 	ctx->pkts_drops += drops;
 	ctx->pkts_recvd_last = packets;
 	ctx->pkts_drops_last = drops;
+	ctx->pkts_skipd_last = 0;
 
 	return 0;
 }
@@ -410,6 +413,27 @@ out:
 	printf("\r%12lu sec, %lu usec in total\n", diff.tv_sec, diff.tv_usec);
 }
 
+static inline bool __skip_packet(struct ctx *ctx, struct sockaddr_ll *sll)
+{
+	if (ctx->packet_type != -1)
+		return ctx->packet_type != sll->sll_pkttype;
+
+	/* when receving from the loopback device, each packet is seen twice,
+	 * so drop the outgoing ones to avoid duplicates
+	 */
+	return (sll->sll_ifindex == ctx->lo_ifindex) &&
+	       (sll->sll_pkttype == PACKET_OUTGOING);
+}
+
+static inline bool skip_packet(struct ctx *ctx, struct sockaddr_ll *sll)
+{
+	bool skip = __skip_packet(ctx, sll);
+
+	if (skip)
+		ctx->pkts_skipd_last++;
+	return skip;
+}
+
 static void receive_to_xmit(struct ctx *ctx)
 {
 	short ifflags = 0;
@@ -468,11 +492,10 @@ static void receive_to_xmit(struct ctx *ctx)
 			hdr_in = rx_ring.frames[it_in].iov_base;
 			in = ((uint8_t *) hdr_in) + hdr_in->tp_h.tp_mac;
 
-			ctx->pkts_seen++;
+			if (skip_packet(ctx, &hdr_in->s_ll))
+				goto next;
 
-			if (ctx->packet_type != -1)
-				if (ctx->packet_type != hdr_in->s_ll.sll_pkttype)
-					goto next;
+			ctx->pkts_seen++;
 
 			hdr_out = tx_ring.frames[it_out].iov_base;
 			out = ((uint8_t *) hdr_out) + TPACKET2_HDRLEN - sizeof(struct sockaddr_ll);
@@ -942,9 +965,8 @@ static void walk_t3_block(struct block_desc *pbd, struct ctx *ctx,
 		uint8_t *packet = ((uint8_t *) hdr + hdr->tp_mac);
 		pcap_pkthdr_t phdr;
 
-		if (ctx->packet_type != -1)
-			if (ctx->packet_type != sll->sll_pkttype)
-				goto next;
+		if (skip_packet(ctx, sll))
+			goto next;
 
 		ctx->pkts_seen++;
 
@@ -1071,9 +1093,8 @@ static void recv_only_or_dump(struct ctx *ctx)
 			uint8_t *packet = ((uint8_t *) hdr) + hdr->tp_h.tp_mac;
 			pcap_pkthdr_t phdr;
 
-			if (ctx->packet_type != -1)
-				if (ctx->packet_type != hdr->s_ll.sll_pkttype)
-					goto next;
+			if (skip_packet(ctx, &hdr->s_ll))
+				goto next;
 
 			ctx->pkts_seen++;
 
@@ -1523,6 +1544,9 @@ int main(int argc, char **argv)
 	if (!ctx.device_in)
 		ctx.device_in = xstrdup("any");
 
+	if (!strcmp(ctx.device_in, "any") || !strcmp(ctx.device_in, "lo"))
+		ctx.lo_ifindex = device_ifindex("lo");
+
 	register_signal(SIGINT, signal_handler);
 	register_signal(SIGQUIT, signal_handler);
 	register_signal(SIGTERM, signal_handler);
