@@ -417,15 +417,9 @@ static struct cpu_stats *setup_shared_var(unsigned int cpus)
 	char *zbuff, file[256];
 	struct cpu_stats *buff;
 
+	slprintf(file, sizeof(file), ".tmp_mmap.XXXXXX");
+	fd = mkostemp_or_die(file, O_RDWR | O_CREAT | O_TRUNC);
 	zbuff = xzmalloc(len);
-	slprintf(file, sizeof(file), ".tmp_mmap.%u", (unsigned int) rand());
-
-	fd = creat(file, S_IRUSR | S_IWUSR);
-	bug_on(fd < 0);
-	close(fd);
-
-	fd = open_or_die_m(file, O_RDWR | O_CREAT | O_TRUNC,
-			   S_IRUSR | S_IWUSR);
 	write_or_die(fd, zbuff, len);
 	xfree(zbuff);
 
@@ -618,13 +612,30 @@ static void shaper_delay(struct shaper *sh, unsigned long pkt_len)
 	}
 }
 
+static inline void packet_apply_dyn_elements(int idx)
+{
+	if (packet_dyn_has_elems(&packet_dyn[idx])) {
+		apply_counter(idx);
+		apply_randomizer(idx);
+		apply_csum16(idx);
+	}
+
+	if (packet_dyn_has_fields(&packet_dyn[idx])) {
+		uint32_t i;
+
+		for (i = 0; i < packet_dyn[idx].flen; i++)
+			proto_field_dyn_apply(packet_dyn[idx].fields[i]);
+
+		proto_packet_update(idx);
+	}
+}
+
 static void xmit_slowpath_or_die(struct ctx *ctx, unsigned int cpu, unsigned long orig_num)
 {
 	int ret, icmp_sock = -1;
 	unsigned long num = 1, i = 0;
 	struct timeval start, end, diff;
 	unsigned long long tx_bytes = 0, tx_packets = 0;
-	struct packet_dyn *pktd;
 	struct sockaddr_ll saddr = {
 		.sll_family = PF_PACKET,
 		.sll_halen = ETH_ALEN,
@@ -647,12 +658,7 @@ static void xmit_slowpath_or_die(struct ctx *ctx, unsigned int cpu, unsigned lon
 		shaper_init(&ctx->sh);
 
 	while (likely(sigint == 0 && num > 0 && plen > 0)) {
-		pktd = &packet_dyn[i];
-		if (packet_dyn_has_elems(pktd)) {
-			apply_counter(i);
-			apply_randomizer(i);
-			apply_csum16(i);
-		}
+		packet_apply_dyn_elements(i);
 retry:
 		ret = sendto(sock, packets[i].payload, packets[i].len, 0,
 			     (struct sockaddr *) &saddr, sizeof(saddr));
@@ -719,7 +725,6 @@ static void xmit_fastpath_or_die(struct ctx *ctx, unsigned int cpu, unsigned lon
 	struct ring tx_ring;
 	struct frame_map *hdr;
 	struct timeval start, end, diff;
-	struct packet_dyn *pktd;
 	unsigned long long tx_bytes = 0, tx_packets = 0;
 
 	set_sock_prio(sock, 512);
@@ -755,12 +760,7 @@ static void xmit_fastpath_or_die(struct ctx *ctx, unsigned int cpu, unsigned lon
 		hdr->tp_h.tp_snaplen = packets[i].len;
 		hdr->tp_h.tp_len = packets[i].len;
 
-		pktd = &packet_dyn[i];
-		if (packet_dyn_has_elems(pktd)) {
-			apply_counter(i);
-			apply_randomizer(i);
-			apply_csum16(i);
-		}
+		packet_apply_dyn_elements(i);
 
 		fmemcpy(out, packets[i].payload, packets[i].len);
 
