@@ -5,6 +5,10 @@
 #include <stdio.h>
 #include <sys/types.h>
 
+#include "trafgen_proto.h"
+
+#define PROTO_MAX_LAYERS	16
+
 #define TYPE_INC	0
 #define TYPE_DEC	1
 
@@ -34,6 +38,8 @@ struct csum16 {
 struct packet {
 	uint8_t *payload;
 	size_t len;
+	struct proto_hdr *headers[PROTO_MAX_LAYERS];
+	size_t headers_count;
 };
 
 struct packet_dyn {
@@ -43,6 +49,8 @@ struct packet_dyn {
 	size_t rlen;
 	struct csum16 *csum;
 	size_t slen;
+	struct proto_field **fields;
+	size_t flen;
 };
 
 static inline bool packet_dyn_has_elems(struct packet_dyn *p)
@@ -55,6 +63,11 @@ static inline bool packet_dyn_has_only_csums(struct packet_dyn *p)
 	return (p->clen == 0 && p->rlen == 0 && p->slen);
 }
 
+static inline bool packet_dyn_has_fields(struct packet_dyn *p)
+{
+	return p->flen;
+}
+
 extern void compile_packets_str(char *str, bool verbose, unsigned int cpu);
 extern void compile_packets(char *file, bool verbose, unsigned int cpu,
 			    bool invoke_cpp, char *const cpp_argv[]);
@@ -63,5 +76,7 @@ extern void cleanup_packets(void);
 extern void set_fill(uint8_t val, size_t len);
 
 extern struct packet *current_packet(void);
+extern uint32_t current_packet_id(void);
+extern struct packet *packet_get(uint32_t id);
 
 #endif /* TRAFGEN_CONF */
